#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

Fuel slug

----------
ATTENTION:
----------
This container requires two Phase() containers which are by definition histories.
The history is not checked. Therefore any inconsistency will be propagated forward.
A fuel slug has two solid phases: cladding and fuel. The user will decide how to
best use the underlying history data in the Phase() container of each phase.

VFdALib support classes

Thu Dec 15 16:18:39 EST 2016
'''
# *******************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random
import pandas
from copy import deepcopy

from cortix.support.periodictable import ELEMENTS
from cortix.support.periodictable import SERIES
from cortix.support.phase import Phase
# *******************************************************************************

class FuelSlug:

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 specs = pandas.Series(),
                 fuelPhase = Phase(),
                 claddingPhase = Phase()
                ):

        assert isinstance(specs, pandas.Series), 'fatal.'
        assert isinstance(fuelPhase, Phase), 'fatal.'
        assert isinstance(claddingPhase, Phase), 'fatal.'

        self.attributeNames = \
            ['nSlugs',
             'slugType',
             'slugVolume',
             'slugArea',
             'fuelVolume',
             'claddingVolume',
             'fuelArea',
             'claddingArea',
             'equivalentCladdingVolume',
             'equivalentCladdingArea',
             'equivalentFuelVolume',
             'equivalentFuelArea',
             'fuelLength',
             'slugLength',
             'fuelMass',
             'fuelMassDens',
             'fuelMassCC',
             'claddingMass',
             'claddingMassDens',
             'claddingMassCC',
             'nuclides',
             'isotopes',
             'radioactivity',
             'radioactivityDens',
             'gamma',
             'gammaDens',
             'heat',
             'heatDens',
             'molarHeatPwr',
             'molarGammaPwr']

        # own internal copy
        self._specs = deepcopy(specs)
        self._fuelPhase = deepcopy(fuelPhase)
        self._claddingPhase = deepcopy(claddingPhase)

        # setup the equivalent cladding hollow sphere
        pi = math.pi

        area = self.__GetAttribute('claddingArea')
        volume = self.__GetAttribute('claddingVolume')

        ro = math.sqrt(area / 4 / pi)
        ri = (ro**3 - volume * 3 / 4 / pi)**(1 / 3)

        self._claddingHollowSphereRo = ro
        self._claddingHollowSphereRi = ri

        area = self.__GetAttribute('fuelArea')
        volume = self.__GetAttribute('fuelVolume')

        ro = math.sqrt(area / 4 / pi)
        ri = (ro**3 - volume * 3 / 4 / pi)**(1 / 3)

        self._fuelHollowSphereRo = ro
        self._fuelHollowSphereRi = ri

        return

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def GetSpecs(self):
        '''
        Returns the species associated with this fuel slug.

        Returns
        -------
        specs: str
        '''

        return self._specs
    specs = property(GetSpecs, None, None, None)

    def GetFuelPhase(self):
        '''
        Returns the phase history of the solid fuel.

        Returns
        -------
        fuelPhase: dataFrame
        '''

        return self._fuelPhase
    fuelPhase = property(GetFuelPhase, None, None, None)

    def GetCladdingPhase(self):
        '''
        Returns the phase history of the cladding.

        Returns
        -------
        claddingPhase: dataFrame
        '''

        return self._claddingPhase
    claddingPhase = property(GetCladdingPhase, None, None, None)

    def GetAttribute(self, name, phase=None, symbol=None, series=None):
        '''
        Returns the value of the specified attribute. Any attribute that is
        specified in class construction can be retrieved using this function.
        The attribute may also be retrived from a speciefic phase, a specific
        nuclide OR a specific series.

        Parameters
        ----------
        name: str
        phase: str
        symbol: str
        series: str

        Returns
        -------
        attribute: int or float
        '''

        return self.__GetAttribute(name, symbol, series)

    def ReduceCladdingVolume(self, dissolvedVolume):
        '''
        Reduces the amount of cladding in the slug by dissolvedvolume.
        This will also update the dimensions of the cladding walls and
        end caps; volume will be taken from all sections equally such that the
        relative dimensions stay the same.

        Parameters
        ----------
        dissolvedVolume: float
        '''

        self.__ReduceCladdingVolume(dissolvedVolume)

    def ReduceFuelVolume(self, dissolvedVolume):
        '''
        Reduces the amount of fuel in the slug by dissolvedVolume. This
        will also update the dimensions of the fuel slug, mainly the thickness
        of each fuel layers.

        Parameters
        ----------
        dissolvedVolume: float
        '''

        self.__ReduceFuelVolume(dissolvedVolume)

    def __str__(self):
        s = 'FuelSlug(): \n\t specs \n\t %s \n\t fuelPhase \n\t %s \n\t claddingPhase \n\t %s'
        return s % (self._specs, self._fuelPhase, self._claddingPhase)

    def __repr__(self):
        s = 'FuelSlug(): \n\t specs \n\t %s \n\t fuelPhase \n\t %s \n\t claddingPhase \n\t %s'
        return s % (self._specs, self._fuelPhase, self._claddingPhase)

# Get stored fuel slug property either overall or on a nuclide basis

#*********************************************************************************
# Private Member Functions (Internal use: __)
#*********************************************************************************

    def __GetAttribute(self, attributeName, nuclide=None, series=None):
        '''
        Returns the specified attribute. Series and nuclide must both be
        specified if the other is specified!

        Parameters
        ----------
        attributeName: str
        nuclide: str
        series: str

        Returns
        -------
        Various attributes.
        '''


        assert attributeName in self.attributeNames, ' attribute name: %r; options: %r; fail.' % (
            attributeName, self.attributeNames)

        if nuclide is not None:
            assert len(nuclide.split('*')) == 1  # sanity check

        if nuclide is not None:
            assert series is None, 'fail.'
        if series is not None:
            assert nuclide is None, 'fail.'

        if series is not None:
            assert False, ' not implemented.'

        if attributeName == 'isotopes':
            assert nuclide is not None, 'need a nuclide symbol.'

# .................................................................................
# # of slugs

        if attributeName == 'nSlugs':
            return 1

# .................................................................................
# slugId

        if attributeName == 'slugType':
            return self._specs['Slug type']

# .................................................................................
# slug volume

        if attributeName == 'slugVolume':
            return __GetSlugVolume(self)

# .................................................................................
# fuel volume

        if attributeName == 'fuelVolume':
            return __GetFuelVolume(self)

# .................................................................................
# equivalent fuel volume

        if attributeName == 'equivalentFuelVolume':
            return __GetEquivalentFuelVolume(self)

# .................................................................................
# cladding volume

        if attributeName == 'claddingVolume':
            return __GetCladdingVolume(self)

# .................................................................................
# equivalent cladding volume

        if attributeName == 'equivalentCladdingVolume':
            return __GetEquivalentCladdingVolume(self)

# .................................................................................
# fuel area

        if attributeName == 'fuelArea':
            return __GetFuelArea(self)

# .................................................................................
# equivalent fuel area

        if attributeName == 'equivalentFuelArea':
            return __GetEquivalentFuelArea(self)

# .................................................................................
# cladding area

        if attributeName == 'claddingArea':
            return __GetCladdingArea(self)
        if attributeName == 'slugArea':
            return __GetCladdingArea(self)

# .................................................................................
# equivalent cladding area

        if attributeName == 'equivalentCladdingArea':
            return __GetEquivalentCladdingArea(self)

# .................................................................................
# fuel length

        if attributeName == 'fuelLength':
            return __GetFuelLength(self)

# .................................................................................
# fuel slug overall quantities
        if nuclide is None and series is None:

            # mass or mass concentration
            if attributeName == 'fuelMassCC' or attributeName == 'fuelMassDens' or attributeName == 'fuelMass':
                mass = self._fuelPhase.GetValue('mass')
                if attributeName == 'fuelMass':
                    return mass
                else:
                    #          volume = __GetFuelVolume( self )
                    volume = __GetEquivalentFuelVolume(self)
                    if volume == 0.0:
                        assert abs(
                            volume - self._fuelPhase.GetValue('volume')) < 1e-8
                        return mass
                    else:
                        assert abs(
                            volume - self._fuelPhase.GetValue('volume')) / volume * 100.0 < 0.1
                        return mass / volume
# mass or mass concentration
            if attributeName == 'claddingMassCC' or attributeName == 'claddingMassDens' or attributeName == 'claddingMass':
                mass = self._claddingPhase.GetValue('mass')
                if attributeName == 'claddingMass':
                    return mass
                else:
                    #          volume = __GetCladdingVolume( self )
                    volume = __GetEquivalentCladdingVolume(self)
                    if volume == 0.0:
                        assert abs(
                            volume - self._claddingPhase.GetValue('volume')) < 1e-8
                        return mass
                    else:
                        assert abs(
                            volume - self._claddingPhase.GetValue('volume')) / volume * 100.0 < 0.1
                        return mass / volume
# radioactivity
            if attributeName == 'radioactivtyDens' or attributeName == 'radioactivity':
                radioactivity = self._fuelPhase.GetValue('radioactivity')
                if attributeName == 'radioactivity':
                    return radioactivity
                else:
                    volume = __GetFuelVolume(self)
                    if volume == 0.0:
                        assert abs(
                            volume - self._fuelPhase.GetValue('volume')) < 1e-8
                        return radioactivity
                    else:
                        assert abs(
                            volume - self._fuelPhase.GetValue('volume')) / volume * 100.0 < 0.1
                        return radioactivity / volume
# gamma
            if attributeName == 'gammaDens' or attributeName == 'gamma':
                gammaDens = 0.0
                for spc in self._fuelPhase.species:
                    gammaDens += spc.molarGammaPwr * spc.molarCC
                if attributeName == 'gammaDens':
                    return gammaDens
                else:
                    volume = __GetFuelVolume(self)
                    return gammaDens * volume
# heat
            if attributeName == 'heatDens' or attributeName == 'heat':
                heatDens = 0.0
                for spc in self._fuelPhase.species:
                    heatDens += spc.molarHeatPwr * spc.molarCC
                if attributeName == 'heatDens':
                    return heatDens
                else:
                    volume = __GetFuelVolume(self)
                    return heatDens * volume

# .................................................................................
# radioactivity

        if attributeName == 'radioactivityDens' or attributeName == 'radioactivity':
            assert False
            colName = 'Radioactivity Dens. [Ci/cc]'

# .................................................................................
# thermal

        if attributeName == 'thermalDens' or attributeName == 'thermal' or  \
           attributeName == 'heatDens' or attributeName == 'heat':
            assert False
            colName = 'Thermal Dens. [W/cc]'

# .................................................................................
# gamma

        if attributeName == 'gammaDens' or attributeName == 'gamma':
            assert False
            colName = 'Gamma Dens. [W/cc]'

# .................................................................................
##########################################################################
# .................................................................................

#  if attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC':
#     attributeDens = True
#  else:
#     attributeDens = False

# .................................................................................
# all nuclide content of the fuel added

#  if nuclide is None and series is None:
#
#     density = 0.0
#
#     density = self.propertyDensities[ colName ].sum()
#
#     if attributeDens is False:
#        volume = __GetFuelSlugVolume( self )
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get chemical element series

#  if series is not None:
#
#     density = 0.0
#
#     for isotope in isotopes:
#       density += self.propertyDensities.loc[isotope,colName]
#
#     if attributeDens is False:
#        volume = __GetFuelSlugVolume( self )
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get specific nuclide (either the isotopes of the nuclide or the specific
# isotope) property

        if nuclide is not None:

            # a particular nuclide given (atomic number and atomic mass number)
            if len(nuclide.split('-')) == 2:

                nuclideMassNumber = int(nuclide.split('-')[1].strip('m'))
                nuclideSymbol = nuclide.split('-')[0]
                nuclideMolarMass = ELEMENTS[nuclideSymbol].isotopes[nuclideMassNumber].mass

                massCC = 0.0

                for spc in self._fuelPhase.species:

                    formula = spc.atoms

                    moleFraction = 0.0

                    for item in formula:

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            formulaNuclideSymbol = item.split('-')[0].strip()
                            if formulaNuclideSymbol == nuclideSymbol:
                                assert len(item.split('-')) == 2

                            if item.split('*')[0].strip() == nuclide:
                                moleFraction = 1.0
                            else:
                                moleFraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            formulaNuclideSymbol = item.split(
                                '*')[1].split('-')[0].strip()
                            if formulaNuclideSymbol == nuclideSymbol:
                                assert len(item.split('*')[1].split('-')) == 2

                            if item.split('*')[1].strip() == nuclide:
                                moleFraction = float(
                                    item.split('*')[0].strip())
                            else:
                                moleFraction = 0.0

                        else:
                            assert False

                        massCC += spc.molarCC * moleFraction * nuclideMolarMass

                return massCC * __GetFuelVolume(self)

        # chemical element given (only atomic number given)
            elif len(nuclide.split('-')) == 1:

                massCC = 0.0

                for spc in self._fuelPhase.species:

                    formula = spc.atoms

                    for item in formula:

                        moleFraction = 0.0

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            assert len(item.split('-')) == 2
                            formulaNuclideSymbol = item.split('-')[0].strip()
                            formulaNuclideMassNumber = int(
                                item.split('-')[1].strip('m'))
                            formulaNuclideMolarMass = ELEMENTS[formulaNuclideSymbol].isotopes[formulaNuclideMassNumber].mass

                            if formulaNuclideSymbol == nuclide:
                                moleFraction = 1.0
                            else:
                                moleFraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            assert len(item.split('*')[1].split('-')) == 2
                            formulaNuclideSymbol = item.split(
                                '*')[1].split('-')[0].strip()
                            formulaNuclideMassNumber = int(
                                item.split('*')[1].split('-')[1].strip('m'))
                            formulaNuclideMolarMass = ELEMENTS[formulaNuclideSymbol].isotopes[formulaNuclideMassNumber].mass

                            if formulaNuclideSymbol == nuclide:
                                moleFraction = float(
                                    item.split('*')[0].strip())
                            else:
                                moleFraction = 0.0

                        else:
                            assert False

                        massCC += spc.molarCC * moleFraction * formulaNuclideMolarMass

                return massCC * __GetFuelVolume(self)

            else:

                assert False

    def __GetSlugLength(self):
        '''
        Returns the length of the fuelslug in cm (includes the end caps).

        Returns
        -------
        self._specs['Slug length [cm]']: float
        '''

        return self._specs['Slug length [cm]']

    def __GetFuelLength(self):
        '''
        Returns the length of the fuel section in the slug (this is the length
        of the fuel slug without the cladding end caps). Given in cm.

        Returns
        -------
        fuelLength: float
        '''

        slugLength = __GetSlugLength(self)

        cladEndCapThickness = self._specs['Cladding end cap thickness [cm]']
        fuelLength = slugLength - 2.0 * cladEndCapThickness

        return fuelLength

    def __GetSlugVolume(self):
        '''
        Returns the total volume of the fuel slug, in cm^3.

        Returns
        -------
        outerSlugVolume + innerSlugVolume: float
        '''

        slugLength = __GetSlugLength(self)

        outerSlugOuterRadius = self._specs['Outer slug OD [cm]'] / 2.0
        outerSlugInnerRadius = self._specs['Outer slug ID [cm]'] / 2.0
        outerSlugVolume = slugLength * math.pi * \
            (outerSlugOuterRadius**2 - outerSlugInnerRadius**2)
        innerSlugOuterRadius = self._specs['Inner slug OD [cm]'] / 2.0
        innerSlugInnerRadius = self._specs['Inner slug ID [cm]'] / 2.0
        innerSlugVolume = slugLength * math.pi * \
            (innerSlugOuterRadius**2 - innerSlugInnerRadius**2)

        return outerSlugVolume + innerSlugVolume

    def __GetFuelVolume(self):
        '''
        Returns the total volume of fuel in the slug, in cm^3.

        Returns
        -------
        volume: float
        '''

        fuelLength = __GetFuelLength(self)

        cladWallThickness = self._specs['Cladding wall thickness [cm]']
        outerFuelOuterRadius = self._specs['Outer slug OD [cm]'] / \
            2.0 - cladWallThickness
        outerFuelInnerRadius = self._specs['Outer slug ID [cm]'] / \
            2.0 + cladWallThickness

        outerFuelVolume = fuelLength * math.pi * \
            (outerFuelOuterRadius**2 - outerFuelInnerRadius**2)

        innerFuelOuterRadius = self._specs['Inner slug OD [cm]'] / \
            2.0 - cladWallThickness
        innerFuelInnerRadius = self._specs['Inner slug ID [cm]'] / \
            2.0 + cladWallThickness

        innerFuelVolume = fuelLength * math.pi * \
            (innerFuelOuterRadius**2 - innerFuelInnerRadius**2)

        volume = outerFuelVolume + innerFuelVolume

        return volume

    def __GetCladdingVolume(self):
        '''
        Returns the total volume of cladding in the slug.

        Returns
        -------
        slugVolume - fuelVolume: float
        '''

        slugVolume = __GetSlugVolume(self)
        fuelVolume = __GetFuelVolume(self)

        return slugVolume - fuelVolume

    def __GetFuelArea(self):
        '''
        Returns the surface area of the fuel, in cm^2.

        Returns
        -------
        outerSlugFuelArea + innerSlugFuelArea: float
        '''

        pi = math.pi

        cladWallThickness = self._specs['Cladding wall thickness [cm]']
        slugLength = self._specs['Slug length [cm]']

        outerSlugFuelArea = 0.0

        # side walls
        outerSlugOuterRadius = self._specs['Outer slug OD [cm]'] / 2.0
        outerSlugFuelArea += 2.0 * pi * \
            (outerSlugOuterRadius - cladWallThickness) * slugLength
        outerSlugInnerRadius = self._specs['Outer slug ID [cm]'] / 2.0
        outerSlugFuelArea += 2.0 * pi * \
            (outerSlugInnerRadius + cladWallThickness) * slugLength
        # add bottom and top areas
        outerSlugFuelArea += 2.0 * pi * \
            ((outerSlugOuterRadius - cladWallThickness)**2 -
             (outerSlugInnerRadius + cladWallThickness)**2)

        innerSlugFuelArea = 0.0

        # side walls
        innerSlugOuterRadius = self._specs['Inner slug OD [cm]'] / 2.0
        innerSlugFuelArea += 2.0 * pi * \
            (innerSlugOuterRadius - cladWallThickness) * slugLength
        innerSlugInnerRadius = self._specs['Inner slug ID [cm]'] / 2.0
        innerSlugFuelArea += 2.0 * pi * \
            (innerSlugInnerRadius + cladWallThickness) * slugLength
        # add bottom and top areas
        innerSlugFuelArea += 2.0 * pi * \
            ((innerSlugOuterRadius - cladWallThickness)**2 -
             (innerSlugInnerRadius + cladWallThickness)**2)

        return outerSlugFuelArea + innerSlugFuelArea

    def __GetCladdingArea(self):
        '''
        Returns the surface area of the cladding, in cm^2.

        Returns
        -------
        outerSlugArea + innerSlugArea: float
        '''

        slugLength = self._specs['Slug length [cm]']

        outerSlugArea = 0.0

        # side walls
        outerSlugOuterRadius = self._specs['Outer slug OD [cm]'] / 2.0
        outerSlugArea += 2.0 * math.pi * outerSlugOuterRadius * slugLength
        outerSlugInnerRadius = self._specs['Outer slug ID [cm]'] / 2.0
        outerSlugArea += 2.0 * math.pi * outerSlugInnerRadius * slugLength
        # add bottom and top areas
        outerSlugArea += 2.0 * math.pi * \
            (outerSlugOuterRadius**2 - outerSlugInnerRadius**2)

        innerSlugArea = 0.0

        # side walls
        innerSlugOuterRadius = self._specs['Inner slug OD [cm]'] / 2.0
        innerSlugArea += 2.0 * math.pi * innerSlugOuterRadius * slugLength
        innerSlugInnerRadius = self._specs['Inner slug ID [cm]'] / 2.0
        innerSlugArea += 2.0 * math.pi * innerSlugInnerRadius * slugLength
        # add bottom and top areas
        innerSlugArea += 2.0 * math.pi * \
            (innerSlugOuterRadius**2 - innerSlugInnerRadius**2)

        return outerSlugArea + innerSlugArea

    def __GetEquivalentCladdingArea(self):
        '''
        Returns the surface area of a sphere with a radius given by
        self._claddingHollowSphereRo with the same surface area as the cladding
        on the fuel slug, in cm^2.

        Returns
        -------
        area: float
        '''

        ro = self._claddingHollowSphereRo

        area = 4.0 * math.pi * ro

        return area

    def __GetEquivalentCladdingVolume(self):
        '''
        Returns the equivalent cladding volume of a hollow sphere that is
        equal to the cladding volume in the fuel slug, in cm^3.

        Returns
        -------
        volume: float
        '''

        ro = self._claddingHollowSphereRo
        ri = self._claddingHollowSphereRi

        volume = 4.0 / 3.0 * math.pi * (ro**3 - ri**3)

        return volume

    def __GetEquivalentFuelVolume(self):
        '''
        This function does the same thing as __GetEquivalentCladdingVolume but
        for fuel instead of clading.

        Returns
        -------
        volume: float
        '''

        ro = self._fuelHollowSphereRo
        ri = self._fuelHollowSphereRi

        volume = 4.0 / 3.0 * math.pi * (ro**3 - ri**3)

        return volume

    def __GetEquivalentFuelArea(self):
        '''
        This function does the same thing as __GetEquiavelntCladdingArea, but
        for fuel instead of cladding.

        Returns
        -------
        area: float
        '''

        ro = self._fuelHollowSphereRo

        area = 4.0 * math.pi * ro

        return area

# Shrink the volume based on the equivalent cladding hollow sphere

    def __ReduceCladdingVolume(self, dissolvedVolume):
        '''
        Reduces the volume of cladding by the specified amount and then shrinks
        the dimensions of the cladding walls to fit the new volume. The
        function does this by first creating a hollow sphere with the same
        volume as the cladding, reduces this sphere's radius until an amount
        of volume specified by dissolvedVolume has been lost, and then
        translates this new volume into new cladding wall dimensions.

        Parameters
        ----------
        dissolvedVolume: float
        '''

        assert dissolvedVolume >= 0.0, 'dissolved volume= %r; failed.' % (
            dissolvedVolume)

        if dissolvedVolume == 0.0:
            return

        if self._claddingHollowSphereRo == self._claddingHollowSphereRi:
            return

        dV = dissolvedVolume
        pi = math.pi

        # get this first
        massDens = _GetAttribute(self, 'claddingMassDens')

# .................................................................................
# reduce the volume of the cladding hollow sphere

        ro = self._claddingHollowSphereRo
        ri = self._claddingHollowSphereRi

        volume = 4.0 / 3.0 * pi * (ro**3 - ri**3)

        if dV < volume:

            ro = (ri**3 + 3.0 / 4.0 / pi * (volume - dV))**(1 / 3)
            self._claddingHollowSphereRo = ro

        else:

            self._claddingHollowSphereRo = 0.0
            self._claddingHollowSphereRi = 0.0

            cladWallThickness = self._specs['Cladding wall thickness [cm]']
            cladEndCapThickness = self._specs['Cladding end cap thickness [cm]']

            self._specs['Inner slug ID [cm]'] += cladWallThickness
            self._specs['Inner slug OD [cm]'] -= cladWallThickness
            self._specs['Outer slug ID [cm]'] += cladWallThickness
            self._specs['Outer slug OD [cm]'] -= cladWallThickness

            self._specs['Slug length [cm]'] -= 2.0 * cladEndCapThickness

            self._specs['Cladding wall thickness [cm]'] = 0.0
            self._specs['Cladding end cap thickness [cm]'] = 0.0

# .................................................................................
# Update the history of the cladding phase

        volume = _GetAttribute(self, 'equivalentCladdingVolume')
        self._claddingPhase.SetValue('volume', volume)

        self._claddingPhase.SetValue('mass', massDens * volume)

    def __ReduceFuelVolume(self, dissolvedVolume):
        '''
        See __ReduceCladdingVolume, above.

        Parameters
        ----------
        dissolvedVolume: float
        '''

        assert dissolvedVolume >= 0.0, 'dissolved volume= %r; failed.' % (
            dissolvedVolume)

        if dissolvedVolume == 0.0:
            return

        if self._fuelHollowSphereRo == self._fuelHollowSphereRi:
            return

        assert self._claddingHollowSphereRo == 0.0
        assert self._claddingHollowSphereRi == 0.0

        dV = dissolvedVolume
        pi = math.pi

        # get this first
        massDens = _GetAttribute(self, 'fuelMassDens')

# .................................................................................
# reduce the volume of the fuel hollow sphere

        ro = self._fuelHollowSphereRo
        ri = self._fuelHollowSphereRi

        volume = 4.0 / 3.0 * pi * (ro**3 - ri**3)

        if dV < volume:

            ro = (ri**3 + 3.0 / 4.0 / pi * (volume - dV))**(1 / 3)
            self._fuelHollowSphereRo = ro

        else:

            self._fuelHollowSphereRo = 0.0
            self._fuelHollowSphereRi = 0.0

            self._specs['Inner slug ID [cm]'] = 0.0
            self._specs['Inner slug OD [cm]'] = 0.0
            self._specs['Outer slug ID [cm]'] = 0.0
            self._specs['Outer slug OD [cm]'] = 0.0

            self._specs['Slug length [cm]'] = 0.0

            self._specs['Cladding wall thickness [cm]'] = 0.0
            self._specs['Cladding end cap thickness [cm]'] = 0.0

# .................................................................................
# Update the history of the fuel phase

        volume = _GetAttribute(self, 'equivalentFuelVolume')
        self._fuelPhase.SetValue('volume', volume)

        self._fuelPhase.SetValue('mass', massDens * volume)

#============================ end class FuelSlug =================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''
Model of a target rod (assembly).
'''
from copy import deepcopy
from cortix.support.phase_new import PhaseNew as Phase

class TargetRod:
    '''Target rod for irradiation and dissolution.

    Two solid phases: cladding and compact
    '''

    def __init__(self, specs=None, cladding_phase=None, compact_phase=None):
        '''Constructor.

        Parameters
        ----------
        specs: dict
        Dictionary of all geometric specifications of the target rod.

        cladding_phase: Phase
        Solid phase for the cladding domain.

        compact_phase: Phase
        Solid phase for the compact domain.

        Attributes
        ----------
        '''

        if specs is not None:
            assert isinstance(specs, dict)
        else:
            self.specs = deepcopy(specs)

        if cladding_phase is not None:
            assert isinstance(cladding_phase, Phase)
        else:
            self.cladding_phase = deepcopy(cladding_phase)

        if compact_phase is not None:
            assert isinstance(compact_phase, Phase)
        else:
            self.compact_phase = deepcopy(compact_phase)

if __name__ == '__main__':
    # Create an empty target rod
    trod = TargetRod()
    print(trod)



import networkx as nx
from cortix.src.module import Module

class Graph:
    def __init__(self):
        self.g = nx.Digraph()

    def add(self, mod):
        assert(isinstance(mod, Module)), "mod must be of type Module!"
        self.g.add_node(mod)
        mod.id = len(self.f.nodes)

    def connect(self, m1, m2):
        assert(isinstance(m1, Module)), "m1 must be of type Module!"
        assert(isinstance(m2, Module)), "m2 must be of type Module!"
        self.g.add_edge(m1, m2)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
"""Crimninal justice network dynamics modeling.

This example uses 7 modules:
    - Community
    - Arrested
    - Adjudication
    - Jail
    - Prison
    - Probation
    - Parole
and a population balance model is used to follow the offenders population
groups between modules.

To run this case using MPI you should compute the number of
processes as follows:

    `nprocs = 7 + 1 cortix`

then issue the MPI run command as follows (replace `nprocs` with a number):

     `mpiexec -n nprocs run_justice.py`

To run this case with the Python multiprocessing library, just run this file at the
command line as

    `run_city_justice.py`

To profile this run do:

     python -m cProfile -s time run_city_justice.py > lixo

"""

import scipy.constants as unit
import matplotlib.pyplot as plt

from cortix import Cortix
from cortix import Network

from cortix.examples.city_justice.prison import Prison
from cortix.examples.city_justice.parole import Parole
from cortix.examples.city_justice.adjudication import Adjudication
from cortix.examples.city_justice.jail import Jail
from cortix.examples.city_justice.arrested import Arrested
from cortix.examples.city_justice.probation import Probation
from cortix.examples.city_justice.community import Community

def main():
    '''Cortix run file for a criminal justice network.

    Attributes
    ----------
    n_groups : int
        Number of population groups being followed. This must be the same for all
        modules.
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    '''

    # Configuration Parameters
    n_groups  =  10 # number of population groups
    end_time  =  30* unit.year
    time_step = 1.0 * unit.day

    use_mpi = False  # True for MPI; False for Python multiprocessing

    city = Cortix(use_mpi=use_mpi, splash=True)

    city.network = Network()

    community = Community(n_groups=n_groups, non_offender_adult_population=100000,
                          offender_pool_size=0)
    city.network.module(community)
    community.end_time = end_time
    community.time_step = time_step
    community.show_time = (True, 30*unit.day)
    community.save = True

    prison = Prison(n_groups=n_groups, pool_size=0.0)
    city.network.module(prison)
    prison.end_time = end_time
    prison.time_step = time_step
    prison.save = True

    parole = Parole(n_groups=n_groups)
    city.network.module(parole)
    parole.end_time = end_time
    parole.time_step = time_step
    parole.save = True

    adjudication = Adjudication(n_groups=n_groups)
    city.network.module(adjudication)
    adjudication.end_time = end_time
    adjudication.time_step = time_step
    adjudication.save = True

    jail = Jail(n_groups=n_groups)
    city.network.module(jail)
    jail.end_time = end_time
    jail.time_step = time_step
    jail.save = True

    arrested = Arrested(n_groups=n_groups)
    city.network.module(arrested)
    arrested.end_time = end_time
    arrested.time_step = time_step
    arrested.save = True

    probation = Probation(n_groups=n_groups)
    city.network.module(probation)
    probation.end_time = end_time
    probation.time_step = time_step
    probation.save = True

    city.network.connect( prison, parole, 'bidirectional' )
    city.network.connect( adjudication, prison )
    city.network.connect( jail, prison )
    city.network.connect( adjudication, jail )
    city.network.connect( arrested, jail )
    city.network.connect( arrested, adjudication )
    city.network.connect( arrested, probation )
    city.network.connect( probation, jail )
    city.network.connect( adjudication, probation )

    city.network.connect( arrested, community, 'bidirectional' )
    city.network.connect( jail, community )
    city.network.connect( probation, community )
    city.network.connect( adjudication, community )
    city.network.connect( prison, community )
    city.network.connect( parole, community )

    city.network.draw()

    city.run()

    if city.use_multiprocessing or city.rank == 0:

        total_num_unknowns = n_groups * len(city.network.modules)
        total_num_params = 0

        # Inspect Data Function
        def inspect_module_data(module, quant_name):
            population_phase = module.population_phase
            (fxg_quant, time_unit) = population_phase.get_quantity_history(quant_name)

            fxg_quant.plot( x_scaling=1/unit.year, x_label='Time [y]',
                    y_label=fxg_quant.latex_name+' ['+fxg_quant.unit+']')

            # Number of parameters in the prison model
            n_params = (len(population_phase.GetActors())-1)*n_groups
            return n_params

        quant_names = {'Prison':'npg','Parole':'feg','Adjudication':'fag',
                'Jail':'fjg','Arrested':'frg','Probation':'fbg','Community':'f0g'}

        for m in city.network.modules:
            quant_name = quant_names[m.name]
            total_num_params += inspect_module_data(m,quant_name)
            plt.grid()
            plt.savefig(m.name+'.png', dpi=300)
            if m.name=='Community':
                inspect_module_data(m,'n0')
                plt.grid()
                plt.savefig(m.name+'-n0.png', dpi=300)

        # Total number of unknowns and parameters

        city.log.info('total number of unknowns   ='+str(total_num_unknowns))
        city.log.info('total number of parameters ='+str(total_num_params))

    # Properly shutdow city
    city.close()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''

'''

#import logging
import scipy.constants as unit
import matplotlib.pyplot as plt

from cortix import Cortix
from cortix import Network

from reactor import BWR
from params import get_params

def main():
    """Cortix run file for a solvent extraction network.

    Attributes
    ----------
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    """

    # Preamble

    end_time = 30 * unit.minute
    time_step = 30.0 # seconds
    show_time = (True, 5*unit.minute)

    use_mpi = False  # True for MPI; False for Python multiprocessing

    # System top level
    plant = Cortix(use_mpi=use_mpi, splash=True)

    # Network
    plant_net = plant.network = Network()

    params = get_params()
    params['start-time'] = 0
    params['end-time'] = end_time

    # Create reactor module
    reactor = BWR(params)

    reactor.name = 'BWR'
    reactor.save = True
    reactor.time_step = time_step
    reactor.end_time = end_time
    reactor.show_time = show_time

    # Add reactor module to network
    plant_net.module(reactor)

    # Create the network connectivity

    #*****************************************************************************

    plant_net.draw()

    # Run network dynamics simulation
    plant.run()

    plot_results = True

    if plot_results and (plant.use_multiprocessing or plant.rank == 0):

        # Reactor graphs
        reactor = plant_net.modules[0]

        (quant, time_unit) = reactor.neutron_phase.get_quantity_history('neutron-dens')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('test-neutron-dens.png', dpi=300)

        (quant, time_unit) = reactor.neutron_phase.get_quantity_history('delayed-neutrons-cc')
        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('test-delayed-neutrons-cc.png', dpi=300)

        (quant, time_unit) = reactor.coolant_outflow_phase.get_quantity_history('temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('test-coolant-outflow-temp.png', dpi=300)

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('fuel-temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label='Fuel Temp. [k]')
        plt.grid()
        plt.savefig('test-fuel-temp.png', dpi=300)

    # Properly shutdown simulation
    plant.close()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from cortix.examples.ideal_gas.particle_plotting import Particle_Plot
from cortix.examples.ideal_gas.particle import Particle
from cortix.examples.ideal_gas.particle_handler import Particle_Handler
from cortix.examples.ideal_gas.run_particle import Run_Particle


##from cortix.examples.city_justice.prison import Prison
##from cortix.examples.city_justice.parole import Parole
##from cortix.examples.city_justice.adjudication import Adjudication
##from cortix.examples.city_justice.jail import Jail
##from cortix.examples.city_justice.arrested import Arrested
##from cortix.examples.city_justice.probation import Probation
##from cortix.examples.city_justice.community import Community




import os, time, datetime, threading, random, sys, string
import numpy as np
from cortix.src.module import Module
from cortix.src.network import Network
from cortix.src.cortix_main import Cortix
from cortix.examples import Particle

class Particle_Handler(Module):

    def __init__(self,shape=None, balls= 5, runtime=3,color='b'):
        super().__init__()
        self.color = color
        self.shape = shape
        self.balls = balls
        self.runtime = runtime
        self.local_balls = []
        self.local_messengers = []
        self.collisions = 0
        self.r = 1
        self.a = (0,0)
        self.cor = 1
        self.t_step = 0.01
        self.timestamp=str(datetime.datetime.now())
        self.elapsed, self.oe = 0,0
        self.name = ''.join([random.choice(string.ascii_lowercase+string.digits) for f in range(10)])
        
    def run(self):
        t = self.t_step
        self.flock = {}
        for i in range(self.balls):
            ball = Particle(shape=self.shape,color=self.color,r=self.r)
            ball.t_step = self.t_step
            ball.a = self.a
            ball.cor=self.cor
            self.local_balls.append(ball)
            self.local_messengers.append(ball.messenger)
            self.flock[ball.name] = ball
        its = round(self.runtime/self.t_step)

        ball_list = []
        
        for messenger in self.local_messengers:
            ball_list.append(messenger)
        for i in self.ports: #Send initial properties
            if 'plot' not in str(i):
                self.send(self.local_messengers,i)
        
        for i in self.ports:
            if 'plot' not in str(i):
                ex_balls = self.recv(i)
                for ball in ex_balls:
                    ball_list.append(ball)
        
        for i in range(its):
            self.elapsed += t
            self.local_messengers = []
            for c,ball in enumerate(ball_list):
                if ball.name in self.flock:
                    messenger = self.flock[ball.name].run(ball_list)
                    ball_list[c] = messenger
                    self.local_messengers.append(messenger)
            ball_list = [f for f in self.local_messengers]
            for i in self.ports: #Send and receive messages for each timestep
                self.send(self.local_messengers,i)
                if 'plot' in str(i):
                    _ = self.recv(i)
                    continue
                messengerlis = self.recv(i)
                for messenger in messengerlis:
                    ball_list.append(messenger)
                

        for i in self.ports: #Send 'done' string to plot module as end condition
            if 'plot' in str(i):
                self.send('Done',i)
        print('Done')
        return



import os, time, datetime, threading, random, sys, string
import numpy as np
from cortix import Module
from cortix import Network
from cortix import Cortix
from cortix.examples import Particle_Plot
from cortix.examples import Particle_Handler

class Run_Particle:
    def __init__(self):
        self.n_list = [15,]

        self.procs = 15
        self.runtime=30
        self.t_step = 0.01
        
        self.r=1
        self.mod_list = []
        self.shape = [(-40, -50), (-40, 90), (90, 60),(60,-30),(30,-50),(-40, -50)]
        self.a = (0,0)
        self.cor = 1
        self.fps = 60

    def run(self):
        if not isinstance(self.n_list, list):
            self.n_list=[self.n_list,]
        for c,i in enumerate(self.n_list):
            self.cortix = Cortix(use_mpi=False)
            self.net = Network()
            self.cortix.network = self.net
            self.plot = Particle_Plot(self.shape,balls=i,modules=self.procs,runtime=self.runtime)
            self.plot.fps = self.fps
            self.net.add_module(self.plot)
            print(c+1,'iterations')
            self.balls = i
            self.balleach = int(self.balls/self.procs)
            remainder = self.balls % self.procs
            self.mod_list = []
            for i in range(self.procs):
                balls = self.balleach
                if remainder > 0:
                    balls+=1
                    remainder -= 1
                app = Particle_Handler(self.shape, balls=balls,runtime = self.runtime)
                app.r=self.r
                app.a=self.a
                app.cor = self.cor
                app.t_step = 0.01
                self.mod_list.append(app)
                self.net.add_module(app)
            for c,i in enumerate(self.mod_list):
                self.net.connect([i,'plot-send{}'.format(c)],[self.plot,'plot-receive{}'.format(c)])
                for j in self.mod_list:
                    if i == j:
                        continue
                    name = '{}{}'.format(i.name,j.name)
                    name2 = '{}{}'.format(j.name,i.name)
                    self.net.connect([i,name], [j,name2])
##            threading.Thread(target=self.plotting,daemon=True).start()
            self.cortix.run()
            self.cortix.close()
            del self.cortix
            print('finished sim')


if __name__ == '__main__':
    sim = Run_Particle()
    sim.runtime = int(input('How many seconds do you want the simulation?: '))
    sim.r = 1
    sim.fps = 100
    sim.shape = [(-40, -50), (-40, 90), (90, 60),(60,-30),(30,-50),(-40, -50)]
##    sim.shape = [(0, 0), (0, 30), (30, 30),(30,25),
##                                  (65,25),(65,-20),(85,-20),(85,-40),
##                                  (45,-40),(45,-20),(49,-20),(49,15),
##                                  (30,15),(30,0),(0,0)]
    sim.t_step = 0.01
    sim.a = (0,0)
    sim.cor = 1
    sim.procs = int(input("How many processes would you like to use?: "))
    sim.n_list = int(input("How many particles would you like to simulate?: "))
    sim.n_list = [sim.n_list,]
    sim.run()




from cortix import Module
from cortix import Port
from time import sleep

class DummyModule(Module):
    def __init__(self):

        # Call the Module class constructor
        super().__init__()

    def run(self):
        # Simulate sending data every second
        i = 0
        while i < 10:
            sleep(1)
            self.send(i, "test-port1")
            print("Sent {}!".format(i))
            i += 1

        print("Finished Sending!")

class DummyModule2(Module):
    def __init__(self):

        # Call the Module class constructor
        super().__init__()

    def run(self):
        # Simulate receiving data every two seconds 
        i = 0
        while i < 10:
            sleep(1)

            # Receive data from all connected ports
            print("Receiving {}!".format(i))
            data = self.recv("test-port2")

            # Extract the data only from test-port1
            print("Received {}!".format(data))

            assert(data == i)
            i += 1

        print("Finished Receiving!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

# Read contents of README file.
from os import path

import setuptools

THIS_DIRECTORY = path.abspath(path.dirname(__file__))

with open(path.join(THIS_DIRECTORY, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

with open('requirements.txt', 'r') as fh:
    REQ = fh.readlines()

setuptools.setup(
    name='cortix',
    version='1.1.33',
    author="Cortix Computing",
    author_email="valmor_dealmeida@uml.edu",
    description='Cortix is a Python library for network dynamics modeling and HPC simulation.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    setup_requires=['wheel'],
    include_package_data=True,
    install_requires=REQ,
    url="https://cortix.org",
    packages=setuptools.find_namespace_packages(),
    keywords = ['simulation', 'math'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Topic :: Scientific/Engineering :: Mathematics',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Topic :: Education',
        'Topic :: Utilities'
    ],
    python_requires='>=3.8'
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
import logging

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix import Phase
from cortix import Quantity

class Community(Module):
    """Community model with criminal group population.

    Community here is the system at large with all possible
    adult individuals included in any given society.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules:

     + `probation`,
     + `adjudication`,
     + `jail`, `prison`,
     + `arrested`, and
     + `parole`.

    See instance attribute `port_names_expected`.

    """

    def __init__(self, n_groups=1, non_offender_adult_population=100,
                 offender_pool_size=0.0, free_offender_pool_size=0.0):
        """Constructor.

        Parameters
        ----------
        n_groups: int
            Number of groups in the population.
        non_offender_adult_population: float
            Pool of individuals reaching the adult age (SI) unit. Default: 100.
        offender_pool_size: float
            Upperbound on the range of the existing population groups. A random value
            from 0 to the upperbound value will be assigned to each group. This is
            typically a small number, say a fraction of a percent.

        """

        super().__init__()

        self.port_names_expected = ['probation', 'adjudication', 'jail', 'prison',
                                    'arrested', 'parole']

        quantities = list()
        self.ode_params = dict()

        self.initial_time = 0.0 * unit.day
        self.end_time = 100 * unit.day
        self.time_step = 0.5 * unit.day
        self.show_time = (False, 10*unit.day)
        self.log = logging.getLogger('cortix')

        unit.percent = 1/100

        # Population groups
        self.n_groups = n_groups

        # Community non-offender population
        n0_0 = np.array([float(non_offender_adult_population)])
        n0 = Quantity(name='n0', formal_name='non-offender-adult-pop',
                      latex_name='$n_0$',
                      unit='# adults', value=n0_0,
                      info='Non-Offender Adult Population')
        quantities.append(n0)

        # Community free-offender population groups
        f0g_0 = np.random.random(self.n_groups) * offender_pool_size
        f0g = Quantity(name='f0g', formal_name='free-offender-pop-grps',
                       latex_name='$n_0^{(g)}$',
                       unit='# offenders', value=f0g_0,
                       info='Free-Offender Population Groups')
        quantities.append(f0g)

        # Model parameters: commitment coefficients

        # Community non-offenders to offenders (arrested)
        a = 0.6*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        c00g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        c00g = Quantity(name='c00g',
                        formal_name='non-offenders-commit-arrested-coeff-grps',
                        unit='1/s', value=c00g_0)
        self.ode_params['non-offenders-commit-to-arrested-coeff-grps'] = c00g_0
        quantities.append(c00g)

        # Community free-offenders to arrested (recidivism)
        a = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.9*unit.percent/unit.year * np.ones(self.n_groups)
        c0rg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        c0rg = Quantity(name='c0rg',
                        formal_name='free-offenders-commit-arrested-coeff-grps',
                        value=c0rg_0, unit='1/s')
        self.ode_params['free-offenders-commit-to-arrested-coeff-grps'] = c0rg_0
        quantities.append(c0rg)

        # Death term for community offenders
        a = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        b = 1.0*unit.percent/unit.year * np.ones(self.n_groups)
        d0g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        self.ode_params['free-offenders-death-rates-coeff'] = d0g_0

        # Death term for community non-offenders
        a = 0.5*unit.percent/unit.year
        b = 1.0*unit.percent/unit.year
        d0_0 = a + (b-a)*np.random.random()
        self.ode_params['non-offenders-death-rate-coeff'] = d0_0

        # Maturity term for community non-offenders
        a = 1.5*unit.percent/unit.year
        b = 2.5*unit.percent/unit.year
        s0_0 = a + (b-a)*np.random.random()
        self.ode_params['non-offenders-maturity-rate-coeff'] = s0_0
        self.ode_params['non_offender_adult_population'] = non_offender_adult_population

        # Phase state
        self.population_phase = Phase(self.initial_time, time_unit='s',
                                      quantities=quantities)

        self.population_phase.SetValue('n0', n0_0, self.initial_time)
        self.population_phase.SetValue('f0g', f0g_0, self.initial_time)

        # Initialize inflows to zero
        self.ode_params['prison-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['parole-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['arrested-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['jail-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['adjudication-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['probation-inflow-rates'] = np.zeros(self.n_groups)

    def run(self, *args):

        self.__zero_ode_parameters()

        time = self.initial_time

        while time <= self.end_time:

            if self.show_time[0] and abs(time%self.show_time[1]-0.0) <= 1.e-1:
                self.log.info('Community::time[d] = '+str(round(time/unit.day, 1)))

            self.__call_ports(time)

            # Evolve offenders group population to the next time stamp
            #---------------------------------------------------------

            time = self.__step(time)

    def __call_ports(self, time):

        # Interactions in the jail port
        #--------------------------------
        # one way "from" jail

        self.send(time, 'jail')
        (check_time, inflow_rates) = self.recv('jail')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['jail-inflow-rates'] = inflow_rates

        # Interactions in the adjudication port
        #--------------------------------------
        # one way "from" adjudication

        self.send(time, 'adjudication')
        (check_time, inflow_rates) = self.recv('adjudication')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['adjudication-inflow-rates'] = inflow_rates

        # Interactions in the probation port
        #--------------------------------
        # one way "from" probation

        self.send(time, 'probation')
        (check_time, inflow_rates) = self.recv('probation')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['probation-inflow-rates'] = inflow_rates

        # Interactions in the prison port
        #--------------------------------
        # one way "from" prison

        self.send(time, 'prison')
        (check_time, inflow_rates) = self.recv('prison')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['prison-inflow-rates'] = inflow_rates

        # Interactions in the parole port
        #--------------------------------
        # one way "from" parole

        self.send(time, 'parole')
        (check_time, inflow_rates) = self.recv('parole')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['parole-inflow-rates'] = inflow_rates

        # Interactions in the arrested port
        #--------------------------------
        # two way "to" and "from" arrested

        # to
        message_time = self.recv('arrested')
        outflow_rates = self.__compute_outflow_rates(message_time, 'arrested')
        self.send((message_time, outflow_rates), 'arrested')

        # from
        self.send(time, 'arrested')
        (check_time, inflow_rates) = self.recv('arrested')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['arrested-inflow-rates'] = inflow_rates

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        # Get state values
        a_vec = self.population_phase.GetValue('n0', time)
        b_vec = self.population_phase.GetValue('f0g', time)

        u_0 = np.concatenate((a_vec, b_vec))

        t_interval_sec = np.linspace(0.0, self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                         u_0, t_interval_sec,
                                         args=(self.ode_params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1, :]  # solution vector at final time step


        time += self.time_step

        # Update state variables
        values = self.population_phase.GetRow() # values existing values
        self.population_phase.AddRow(time, values) # copy on new time for convenience

        self.population_phase.SetValue('n0', u_vec[:1], time) # insert new values
        self.population_phase.SetValue('f0g', u_vec[1:], time) # insert new values

        return time

    def __rhs_fn(self, u_vec, t, params):

        n0 = u_vec[:1]   # non-offender population

        # source of non-offenders
        s0_coeff = params['non-offenders-maturity-rate-coeff']
        n0_0 = params['non_offender_adult_population']
        s0 = s0_coeff * n0_0

        # outflow rate to Arrested
        c00g = params['non-offenders-commit-to-arrested-coeff-grps']
        outflow_rate = np.sum(c00g * n0)

        # death rate
        death_rate_coeff = params['non-offenders-death-rate-coeff']
        death_rate = death_rate_coeff * n0

        dt_n0 = s0 - outflow_rate - death_rate


        f0g = u_vec[1:]  # free-offender population groups

        prison_inflow_rates = params['prison-inflow-rates']
        parole_inflow_rates = params['parole-inflow-rates']
        arrested_inflow_rates = params['arrested-inflow-rates']
        jail_inflow_rates = params['jail-inflow-rates']
        adjudication_inflow_rates = params['adjudication-inflow-rates']
        probation_inflow_rates = params['probation-inflow-rates']

        # free-offenders inflows
        inflow_rates = prison_inflow_rates + parole_inflow_rates +\
                       arrested_inflow_rates + jail_inflow_rates +\
                       adjudication_inflow_rates + probation_inflow_rates

        assert np.all(inflow_rates >= 0.0), 'values: %r'%inflow_rates

        c0rg = params['free-offenders-commit-to-arrested-coeff-grps']


        # free-offenders outflow (recidivism)
        outflow_rates = c0rg * f0g

        assert np.all(outflow_rates >= 0.0), 'values: %r'%outflow_rates

        death_rates_coeff = params['free-offenders-death-rates-coeff']

        death_rates = death_rates_coeff * f0g

        assert np.all(death_rates >= 0.0), 'values: %r'%death_rates

        dt_f0g = inflow_rates - outflow_rates - death_rates


        dt_u = np.concatenate((dt_n0, dt_f0g))

        return dt_u

    def __compute_outflow_rates(self, time, name):

        n0 = self.population_phase.GetValue('n0', time)
        f0g = self.population_phase.GetValue('f0g', time)

        if name == 'arrested':

            c0rg = self.ode_params['free-offenders-commit-to-arrested-coeff-grps']

            c00g = self.ode_params['non-offenders-commit-to-arrested-coeff-grps']

            # Recidivism and new offenders
            outflow_rates = c0rg * f0g + c00g * n0

        return outflow_rates

    def __zero_ode_parameters(self):
        '''
        If ports are not connected the corresponding outflows must be zero.

        '''

        zeros = np.zeros(self.n_groups)

        p_names = [p.name for p in self.ports]

        if 'arrested' not in p_names:
            self.ode_params['commit-to-arrested-coeff-grps'] = zeros

            self.ode_params['general-commit-to-arrested-coeff-grps'] = zeros

        return
